"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const default_1 = require("../config/default");
const util_1 = require("./util");
const capturer_1 = require("./capturer");
const noColor = {
    open: '',
    close: ''
};
let globalConfig;
if (!globalConfig)
    globalConfig = loadConfig();
function loadConfig(newConfig = null) {
    const config = newConfig || default_1.default;
    if (!config.formatters[config.outputFormat])
        throw new Error('There is no formatter for chosen output format');
    return config;
}
function configure(config) {
    globalConfig = loadConfig(config);
}
exports.configure = configure;
exports.context = (activeContext) => usingConfig(() => required(globalConfig)).context(activeContext);
function getGlobalConfig() {
    return globalConfig;
}
exports.getGlobalConfig = getGlobalConfig;
function usingConfig(configSource) {
    const config = typeof configSource === 'function' ? configSource() : configSource;
    return {
        context: forContext,
        reconfigure,
        getConfig: () => config
    };
    function forContext(activeContext) {
        const impl = {};
        for (const type of config.logLevels) {
            impl[type] = (...toLog) => logEvent(config, activeContext, type, toLog);
            const typeLower = type.toLowerCase();
            if (!impl[typeLower]) {
                impl[typeLower] = (...toLog) => logEvent(config, activeContext, type, toLog);
            }
        }
        return impl;
    }
    function reconfigure(newConfig) {
        if (newConfig.logLevels && !util_1.logLevelsMatch(config.logLevels, newConfig.logLevels)) {
            throw new Error('Log levels cannot be updated with reconfigure');
        }
        Object.assign(config, newConfig);
    }
}
exports.usingConfig = usingConfig;
function logEvent(config, activeContext, logLevel, toLog) {
    const eventLogPrio = config.logLevels.indexOf(logLevel);
    for (const transportName of Object.keys(config.transports)) {
        const transport = config.transports[transportName];
        const activeLogLevelPrio = config.logLevels.indexOf(getLogLevel(config, transport, activeContext));
        if (eventLogPrio < activeLogLevelPrio)
            continue;
        const colorEnabled = isColorEnabled(config, transport), color = colorEnabled ? getConfigColor(config, transport, logLevel) || noColor : noColor, openColor = asArray(color).map(c => c.open).join(''), closeColor = asArray(color).reverse().map(c => c.close).join(''), outputMapper = transport.outputMapper || config.outputMapper, mappedValues = toLog.map((value, i) => outputMapper(value, logLevel, activeContext, toLog, i)), outputImpl = transport.outputImpl[logLevel] || transport.outputImpl.default;
        if (!outputImpl)
            throw new Error('Transport ' + transportName + ' does not support log level ' + logLevel + ' or default');
        const outputFormat = transport.outputFormat || config.outputFormat;
        const formatter = (transport.formatters || {})[outputFormat] || config.formatters[outputFormat];
        const formatted = formatter(mappedValues, logLevel, activeContext, config, openColor, closeColor, transport);
        capturer_1.toggleForceCaptureDisabled(true);
        outputImpl(formatted, logLevel, activeContext, config);
        capturer_1.toggleForceCaptureDisabled(false);
    }
}
function getConfigColor(config, transport, logLevel) {
    const colors = transport.colors || config.colors;
    return colors[logLevel] || colors.default;
}
function getLogLevel(config, transport, activeContext) {
    return getOverrideLogLevel(config, activeContext) || getContextSpecificOverrideFromContext() || getOverrideLogLevel(config, null) || transport.defaultLogLevel || config.defaultLogLevel;
    function getContextSpecificOverrideFromContext() {
        if (!(config.contextSpecificLogLevels instanceof Map)) {
            return config.contextSpecificLogLevels[activeContext];
        }
        const perfectOverride = config.contextSpecificLogLevels.get(activeContext);
        if (perfectOverride)
            return perfectOverride;
        for (const key of Array.from(config.contextSpecificLogLevels.keys())) {
            if (key instanceof RegExp && key.test(activeContext)) {
                return config.contextSpecificLogLevels.get(key);
            }
        }
        return perfectOverride;
    }
}
function getOverrideLogLevel(config, activeContext) {
    const prefix = config.envVariablePrefix + '_LOG_LEVEL';
    return process.env[!activeContext ? prefix : prefix + '_' + toEnv(activeContext)];
}
function toEnv(activeContext) {
    return activeContext.replace(/[^a-zA-Z0-9]+/g, '_');
}
function required(val) {
    if (!val)
        throw new Error('Required value missing');
    return val;
}
function asArray(inVal) {
    if (inVal instanceof Array)
        return inVal;
    return [inVal];
}
function isColorEnabled(config, transport) {
    const override = process.env[config.envVariablePrefix + '_COLOR'];
    if (override === '0' || override === 'false')
        return false;
    if (override === '1' || override === 'true')
        return true;
    return typeof transport.supportsColor === 'function' ? transport.supportsColor() : transport.supportsColor;
}
exports.captureStandardOutput = (logger, stdoutLogLevel = 'INFO', stderrLogLevel = 'ERROR') => {
    if (!logger)
        throw new Error('Please provide a logger in the proper context to captureStandardOutput');
    return capturer_1.enableStandardOutputCapture(logger, stdoutLogLevel, stderrLogLevel);
};
